use rumqttc::{MqttOptions, Client, QoS, LastWill};
use bytes::Bytes;
use std::time::Duration;
use std::thread;
use std::sync::mpsc;
use std::collections::hash_map::HashMap;
use prost::Message;
use std::io::Cursor;
use serde::{Serialize, Deserialize};

use std::io::prelude::*;

use sha2::Digest;
use sha2::Sha256;

use std::net::TcpStream;
use std::sync::Mutex;
use std::sync::Arc;

//  MQTT Topics:
//   - division/{division_id}
//   - division/{division_id}/ranking
//   - arena/{arena_id}/score
//   - arena/{arena_id}/state
//   - arena/{arena_id}
//   - game/{division_id}/{game_id}/score
//   - team/{team_string}

pub mod tm {
    include!(concat!(env!("OUT_DIR"), "/tm.rs"));
}

pub fn deserialize_notice(buf: &[u8]) -> Result<tm::Notice, prost::DecodeError> {
    tm::Notice::decode(&mut Cursor::new(buf))
}

#[derive(Serialize, Deserialize, Debug)]
struct DivisionInfo {
    arena: String,
    game_id: String,
}

#[derive(Serialize, Deserialize, Debug)]
struct DivisionRankingInfo {
    rankings: Vec<String>,
}

#[derive(Serialize, Deserialize, Debug)]
enum GameSide {
    Red,
    Blue,
}

#[derive(Serialize, Deserialize, Debug)]
enum ElevationTier {
    A = 0,
    B = 1,
    C = 2,
    D = 3,
    E = 4,
    F = 5,
    G = 6,
    H = 7,
    I = 8,
    J = 9,
}

#[derive(Serialize, Deserialize, Debug)]
struct AllianceScore {
    team_goal: usize,
    team_zone: usize,
    green_goal: usize,
    green_zone: usize,
    elevation_tiers: [Option<ElevationTier>; 2],
}

#[derive(Serialize, Deserialize, Debug)]
struct GameScore {
    autonomous_winner: Option<GameSide>,
    red_score: AllianceScore,
    red_total: usize,
    blue_score: AllianceScore,
    blue_total: usize,
}

#[derive(Serialize, Deserialize, Debug)]
enum GameState {
    Scheduled,
    Timeout,
    Driver,
    Driverdone,
    Autonomous,
    AutonomousDone,
    Abandoned,
}

#[derive(Serialize, Deserialize, Debug)]
struct ArenaStateInfo {
    state: Option<GameState>,
    start_s: usize,
    start_ns: usize,
}

#[derive(Serialize, Deserialize, Debug)]
enum Round {
    None = 0,
    Practice = 1,
    Qualification = 2,
    QuarterFinals = 3,
    SemiFinals = 4,
    Finals = 5,
    RoundOf16 = 6,
    RoundOf32 = 7,
    RoundOf64 = 8,
    RoundOf128 = 9,
    TopN = 15,
    RoundRobin = 16,
    PreEliminations = 20,
    Eliminations = 21,
}

#[derive(Serialize, Deserialize, Debug)]
struct MatchTuple {
    division: String,
    round: Round,
    instance: usize,
    match_num: usize,
    session: usize,
}

#[derive(Serialize, Deserialize, Debug)]
struct ArenaInfo {
    red_teams: [String; 2],
    blue_teams: [String; 2],
    match_tuple: MatchTuple,
}

struct MQTTMessage {
    topic: String,
    payload: String,
}

struct Event {
    name: String,
    divisions: Vec<Division>,
}

struct Division {
    name: String,
    matches: Vec<Match>,
    field_set: FieldSet,
}

struct FieldSet {
    fields: Vec<Field>,
}

struct Field {
    name: String,
    current_match: u32,
}

struct Match {
    name: String,
}

struct TMClient {
    stream: Arc<Mutex<openssl::ssl::SslStream<TcpStream>>>,
    notices: mpsc::Sender<tm::Notice>,
    responses: mpsc::Sender<BackendMessage>,
}

struct TMConnection {
    stream: Arc<Mutex<openssl::ssl::SslStream<TcpStream>>>,
    notices: mpsc::Receiver<tm::Notice>,
    responses: mpsc::Receiver<BackendMessage>,
}

struct BackendMessage {
    status: u8,
    request_id: u32,
    data: Vec<u8>,
}

impl BackendMessage {
    fn from_bytes(bytes: Vec<u8>) -> Option<BackendMessage> {
        if bytes.len() < 5 {
            return None;
        }

        return Some(BackendMessage{
            status: bytes[0],
            request_id: u32::from_le_bytes(bytes[0..4].try_into().unwrap()),
            data: bytes[5..].to_vec(),
        });
    }

    fn as_bytes(self: BackendMessage) -> Vec<u8> {
        let mut bytes = Vec::new();

        bytes.push(self.status);
        bytes.extend(self.request_id.to_le_bytes());
        bytes.extend(self.data);

        return bytes;
    }
}

const BACKEND_PACKET_HEADER_SIZE: usize = 28;
#[derive(Debug)]
struct BackendPacket {
    header: u32,
    timestamp: f64,
    msg_type: u32,
    seq_num: u64,
    size: u32,
    data: Vec<u8>,
}

impl BackendPacket {
    fn new(header: u32, timestamp: f64, msg_type: u32, seq_num: u64, data: Vec<u8>) -> BackendPacket {
        return BackendPacket{
            header,
            timestamp,
            msg_type,
            seq_num,
            size: data.len().try_into().unwrap(),
            data,
        };
    }

    fn from_bytes(bytes: Vec<u8>) -> Option<BackendPacket> {
        if bytes.len() < BACKEND_PACKET_HEADER_SIZE {
            return None;
        }

        return Some(BackendPacket{
            header: u32::from_le_bytes(bytes[0..4].try_into().unwrap()),
            timestamp: f64::from_le_bytes(bytes[4..12].try_into().unwrap()),
            msg_type: u32::from_le_bytes(bytes[12..16].try_into().unwrap()),
            seq_num: u64::from_le_bytes(bytes[16..24].try_into().unwrap()),
            size: u32::from_le_bytes(bytes[24..28].try_into().unwrap()),
            data: bytes[28..].to_vec(),
        });
    }

    fn as_bytes(self: &BackendPacket) -> Vec<u8> {
        let mut bytes = Vec::new();

        bytes.extend(self.header.to_le_bytes());
        bytes.extend(self.timestamp.to_le_bytes());
        bytes.extend(self.msg_type.to_le_bytes());
        bytes.extend(self.seq_num.to_le_bytes());
        bytes.extend(self.size.to_le_bytes());
        bytes.extend(self.data.clone());

        return bytes;
    }
}

const CONNECT_MSG_LEN: usize = 114;
#[derive(Debug)]
struct ConnectMsg {
    version: u32,
    uuid: [u8; 16],
    last_notice_id: u64,
    username: [u8; 16],
    pass_hash: [u8; 32],
    pw_valid: u8,
    state_valid: u8,
    client_name: [u8; 32],
    server_time_zone: i32,
}

impl ConnectMsg {
    fn from_welcome(welcome: ConnectMsg, password: &str, uuid: [u8; 16], client_name: [u8; 32]) -> ConnectMsg {
        let mut hasher = Sha256::new();
        hasher.update(welcome.pass_hash);
        hasher.update(password);
        let result = hasher.finalize();

        return ConnectMsg{
            version: welcome.version,
            uuid,
            last_notice_id: welcome.last_notice_id,
            username: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            pass_hash: result.try_into().unwrap(),
            pw_valid: welcome.pw_valid,
            state_valid: welcome.state_valid,
            client_name,
            server_time_zone: welcome.server_time_zone,
        };
    }

    fn from_bytes(bytes: Vec<u8>) -> Option<ConnectMsg> {
        if bytes.len() < CONNECT_MSG_LEN {
            return None;
        }

        return Some(ConnectMsg{
            version: u32::from_le_bytes(bytes[0..4].try_into().unwrap()),
            uuid: bytes[4..20].to_owned().try_into().unwrap(),
            last_notice_id: u64::from_le_bytes(bytes[20..28].try_into().unwrap()),
            username: bytes[28..44].to_owned().try_into().unwrap(),
            pass_hash: bytes[44..76].to_owned().try_into().unwrap(),
            pw_valid: bytes[77].to_owned(),
            state_valid: bytes[78].to_owned(),
            client_name: bytes[78..110].to_owned().try_into().unwrap(),
            server_time_zone: i32::from_le_bytes(bytes[110..114].try_into().unwrap()),
        });
    }

    fn as_bytes(self: &ConnectMsg) -> Vec<u8> {
        let mut bytes = Vec::new();

        bytes.extend(self.version.to_le_bytes());
        bytes.extend(self.uuid);
        bytes.extend(self.last_notice_id.to_le_bytes());
        bytes.extend(self.username);
        bytes.extend(self.pass_hash);
        bytes.extend(self.pw_valid.to_le_bytes());
        bytes.extend(self.state_valid.to_le_bytes());
        bytes.extend(self.client_name);
        bytes.extend(self.server_time_zone.to_le_bytes());

        return bytes;
    }
}

impl TMClient {
    fn new() -> (TMClient, TMConnection) {
        let (notice_tx, notice_rx) = mpsc::channel();
        let (response_tx, response_rx) = mpsc::channel();

        let mut builder = openssl::ssl::SslConnector::builder(openssl::ssl::SslMethod::tls()).unwrap();
        builder.set_ca_file("tm.crt").unwrap();
        builder.set_verify(openssl::ssl::SslVerifyMode::PEER);

        let connector = builder.build();

        let stream = TcpStream::connect("127.0.0.1:5000").unwrap();

        let mut stream_config = connector.configure().unwrap();
        stream_config.set_verify_hostname(false);
        stream_config.set_certificate_chain_file("tm.crt").unwrap();
        stream_config.set_private_key_file("tm.crt", openssl::ssl::SslFiletype::PEM).unwrap();
        stream_config.set_use_server_name_indication(false);

        let stream = stream_config.connect("127.0.0.1", stream).unwrap();
        let stream_arc = Arc::new(Mutex::new(stream));

        return (TMClient{
            stream: stream_arc.clone(),
            notices: notice_tx,
            responses: response_tx
        },
        TMConnection{
            stream: stream_arc,
            notices: notice_rx,
            responses: response_rx
        },);
    }

    fn process(self: &TMClient) {

        let mut incoming = [0; 2048];
        let mut stream = self.stream.lock().unwrap();
        match stream.read(&mut incoming) {
            Ok(read) => {
                let data = incoming[0..read].to_vec();
                match BackendPacket::from_bytes(data) {
                    Some(packet) => {
                        match packet.msg_type {
                            2 => {
                                match ConnectMsg::from_bytes(packet.data) {
                                    Some(welcome_msg) => {
                                        println!("Welcome msg: {:?}", welcome_msg);
                                        if welcome_msg.pw_valid == 0 {
                                            let uuid = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16];
                                            let client_name = [116, 101, 115, 116, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
                                            let connect_response = ConnectMsg::from_welcome(welcome_msg, "", uuid, client_name);
                                            let response = BackendPacket::new(packet.header, packet.timestamp, packet.msg_type, packet.seq_num, connect_response.as_bytes());
                                            match stream.write(&response.as_bytes()) {
                                                Err(error) => println!("Send error: {:?}", error),
                                                Ok(sent) => println!("Sent {} bytes", sent),
                                            }
                                        } else {
                                            println!("Connected to TM backend!");
                                        }
                                    },
                                    None => println!("Failed to parse welcome msg"),
                                }
                            },
                            _ => println!("Unhandled message type: {}", packet.msg_type),
                        }
                    },
                    None => {
                        panic!("Failed to parse BackendPacket({}): {}", read, String::from_utf8_lossy(&incoming));
                    }
                }
            },
            Err(error) => println!("Error: {}", error),
        }

    }
}

impl TMConnection {
    fn send(self: &mut TMConnection, request: BackendMessage) -> Result<BackendMessage, std::io::Error> {
        let mut stream = self.stream.lock().unwrap();
        match stream.write(&request.as_bytes()) {
            Ok(_) =>
                match self.responses.recv() {
                    Ok(response) => Ok(response),
                    Err(_) => Err(std::io::Error::new(std::io::ErrorKind::ConnectionAborted, "channel closed")),
                },
            Err(error) => Err(error),
        }
    }
}

type NoticeCallback = fn(tm::Notice, Event, ) -> (Vec<MQTTMessage>, Event);

fn get_game_score(scores: tm::MatchScore) -> Option<GameScore> {
    if scores.alliances.len() != 2 {
        return None;
    }

    let ref red_score = scores.alliances[0];
    let ref blue_score = scores.alliances[1];

    // 1) Get the autonomous winner
    // 2) Get score object and fill AllianceScore struct
    // 3) Compute total scores

    let out = GameScore{
        autonomous_winner: None,
        red_total: 0,
        blue_total: 0,
        blue_score: AllianceScore{
            team_goal: 0,
            team_zone: 0,
            green_goal: 0,
            green_zone: 0,
            elevation_tiers: [None, None],
        },
        red_score : AllianceScore{
            team_goal: 0,
            team_zone: 0,
            green_goal: 0,
            green_zone: 0,
            elevation_tiers: [None, None],
        },
    };

    return Some(out);
}

fn on_score_change(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    match notice.match_score {
        None => return (Vec::new(), event),
        Some(game_scores) => {
            match get_game_score(game_scores) {
                Some(score_json) => {
                    let serialized = serde_json::to_string(&score_json).unwrap();
                    let arena_topic = String::from("arena/TEST/score");
                    let mut out = Vec::new();
                    out.push(MQTTMessage{
                        topic: arena_topic,
                        payload: serialized,
                    });
                    return (out, event);
                },
                None => return (Vec::new(), event),
            }
        },
    }
}

fn on_match_start(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    return (Vec::new(), event);
}

fn on_match_cancel(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    return (Vec::new(), event);
}

fn on_match_reset(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    return (Vec::new(), event);
}

fn on_match_assigned(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    return (Vec::new(), event);
}

fn on_active_field_changed(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    return (Vec::new(), event);
}

fn on_rankings_updated(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    return (Vec::new(), event);
}

fn on_event_status_updated(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    return (Vec::new(), event);
}

fn on_elim_alliance_update(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    return (Vec::new(), event);
}

fn on_elim_unavail_teams_update(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    return (Vec::new(), event);
}

fn on_match_list_update(notice: tm::Notice, event: Event) -> (Vec<MQTTMessage>, Event) {
    return (Vec::new(), event);
}

fn main() {
    env_logger::init();

    let mut event = Event{
        name: String::from(""),
        divisions: Vec::new(),
    };

    let mut callbacks: HashMap<tm::NoticeId, NoticeCallback> = HashMap::new();
    callbacks.insert(tm::NoticeId::NoticeRealtimeScoreChanged, on_score_change);
    callbacks.insert(tm::NoticeId::NoticeFieldTimerStarted, on_match_start);
    callbacks.insert(tm::NoticeId::NoticeFieldTimerStopped, on_match_cancel);
    callbacks.insert(tm::NoticeId::NoticeFieldResetTimer, on_match_reset);
    callbacks.insert(tm::NoticeId::NoticeFieldMatchAssigned, on_match_assigned);
    callbacks.insert(tm::NoticeId::NoticeActiveFieldChanged, on_active_field_changed);
    callbacks.insert(tm::NoticeId::NoticeRankingsUpdated, on_rankings_updated);
    callbacks.insert(tm::NoticeId::NoticeEventStatusUpdated, on_event_status_updated);
    callbacks.insert(tm::NoticeId::NoticeElimAllianceUpdated, on_elim_alliance_update);
    callbacks.insert(tm::NoticeId::NoticeElimUnavailTeamsUpdated, on_elim_unavail_teams_update);
    callbacks.insert(tm::NoticeId::NoticeMatchListUpdated, on_match_list_update);

    let mut mqttoptions = MqttOptions::new("vex-bridge", "localhost", 1883);
    mqttoptions.set_keep_alive(Duration::from_secs(5));
    mqttoptions.set_last_will(LastWill{
        topic: String::from("bridge/status"),
        message: Bytes::from("{\"online\": false}"),
        qos: QoS::AtLeastOnce,
        retain: true,
    });

    let (mut client, mut connection) = Client::new(mqttoptions, 10);
    client.subscribe("bridge", QoS::AtLeastOnce).unwrap();
    client.publish("bridge/status", QoS::AtLeastOnce, true, "{\"online\": true}").unwrap();

    let mqtt_recv_thread = thread::spawn(move ||
        for _ in connection.iter() {
        }
    );

    let running = true;
    let (tm_client, tm_connection) = TMClient::new();
    let tm_thread = thread::spawn(move ||
        while running {
            tm_client.process();
        }
    );

    while running {
        thread::sleep(Duration::from_millis(1000));
        match tm_connection.notices.recv() {
            Ok(notice) => {
                let callback = callbacks.get(&notice.id());
                match callback {
                    None => {
                        match notice.id {
                            None => println!("Notice without NoticeId received"),
                            Some(notice_id) => println!("Unhandled NoticeId: {}", notice_id),
                        }
                    },
                    Some(callback) => {
                        let (messages, next_event) = callback(notice, event);
                        event = next_event;
                        for message in messages {
                            let result = client.publish(message.topic, QoS::AtMostOnce, true, message.payload);
                            match result {
                                Ok(_) => {},
                                Err(error) => println!("Publish error: {}", error),
                            }
                        }
                    },
                }
            },
            Err(error) => println!("Notice recv error: {}", error),
        }

    }

    mqtt_recv_thread.join().expect("Failed to join mqtt thread");
    tm_thread.join().expect("Failed to join tm connection thread");
}
